/* packet-zepv3.c
 * Dissector  routines for the ZigBee Encapsulation Protocol
 * By Owen Kirby <osk@exegin.com>, ZEPv3 added by Martin Leixner <info@sewio.net> Sewio Networks
 * Copyright 2009 Exegin Technologies Limited
 *
 * $Id: packet-zepv3.c 47390 2013-01-31 18:31:28Z wmeier $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *------------------------------------------------------------
 *
 *      ZEP Packets must be received in the following format:
 *      |UDP Header|  ZEP Header |IEEE 802.15.4 Packet|
 *      | 8 bytes  | 16/32 bytes |    <= 127 bytes    |
 *------------------------------------------------------------
 *
 *      ZEP v1 Header will have the following format:
 *      |Preamble|Version|Channel ID|Device ID|CRC/LQI Mode|LQI Val|Reserved|Length|
 *      |2 bytes |1 byte |  1 byte  | 2 bytes |   1 byte   |1 byte |7 bytes |1 byte|
 *
 *      ZEP v2 Header will have the following format (if type=1/Data):
 *      |Preamble|Version| Type |Channel ID|Device ID|CRC/LQI Mode|LQI Val|NTP Timestamp|Sequence#|Reserved|Length|
 *      |2 bytes |1 byte |1 byte|  1 byte  | 2 bytes |   1 byte   |1 byte |   8 bytes   | 4 bytes |10 bytes|1 byte|
 *
 *      ZEP v2 Header will have the following format (if type=2/Ack):
 *      |Preamble|Version| Type |Sequence#|
 *      |2 bytes |1 byte |1 byte| 4 bytes |
 *     
 *      ZEP v3 Header will have the following format (if type=1/Data):
 *      |Preamble|Version| Type |Channel ID|Device ID|CRC/LQI Mode|LQI Val|Relative Timestamp|Sequence#| Band |Channel page|Reserved|Length|
 *      |2 bytes |1 byte |1 byte|  1 byte  | 2 bytes |   1 byte   |1 byte |     8 bytes      | 4 bytes |1 byte|   1 byte   | 8 bytes|1 byte|
 * 
 *      ZEP v3 Header will have the following format (if type=2/Ack):
 *      |Preamble|Version| Type |Sequence#|
 *      |2 bytes |1 byte |1 byte| 4 bytes | 
 *------------------------------------------------------------
 */

#include "config.h"

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#include <sys/stat.h>

#include <epan/conversation.h>
#include <epan/emem.h>

#include <epan/nstime.h>

#include <string.h>     
#include <glib.h>

#include <epan/packet.h>
#include <epan/prefs.h>

//#include "packet-ntp.h"
//#include <epan/dissectors/packet-ntp.h>

#include "packet-zepv3.h"


  
/*  Function declarations */
void proto_reg_handoff_zepv3(void);

/*  Initialize protocol and registered fields. */
static int proto_zepv3 = -1;
static int hf_zepv3_version = -1;
static int hf_zepv3_type = -1; 
static int hf_zepv3_channel_id = -1;
static int hf_zepv3_device_id = -1;
static int hf_zepv3_lqi_mode = -1;
static int hf_zepv3_lqi = -1;
static int hf_zepv3_timestamp = -1; 
static int hf_zepv3_snifftimestamp = -1;
static int hf_zepv3_reltimestamp = -1;
static int hf_zepv3_diftimestamp = -1;   
static int hf_zepv3_seqno = -1;
static int hf_zepv3_band = -1;
static int hf_zepv3_chanpage = -1;

static int hf_zepv3_ieee_length = -1;

/* Initialize protocol subtrees. */
static gint ett_zepv3 = -1;
//static gint ett_time_zepv3 = -1;

/* Initialize preferences. */
static guint32  gPREF_zepv3_udp_port = ZEP_DEFAULT_PORT;

/*  Dissector handles */
static dissector_handle_t data_handle;
static dissector_handle_t ieee802154_handle;
static dissector_handle_t ieee802154_ccfcs_handle;


static const value_string bandstrings[] = {
    { 1, "780 MHz" },
    { 2, "868 MHz" },
    { 3, "915 MHz" },
    { 4, "2400 MHz" },
    { 0,       NULL }
};

/* NTP_BASETIME is in fact epoch - ntp_start_time */
#define NTP_BASETIME 2208988800ul
#define NTP_FLOAT_DENOM 4294967296.0

 /*FUNCTION:------------------------------------------------------*/
void
ntp_to_nstime(tvbuff_t *tvb, gint offset, nstime_t *nstime)
{
	guint32		 tempstmp;

	/* We need a temporary variable here so the unsigned math
	 * works correctly (for years > 2036 according to RFC 2030
	 * chapter 3).
	 */
	tempstmp  = tvb_get_ntohl(tvb, offset);
	if (tempstmp)
		nstime->secs = tempstmp - (guint32)NTP_BASETIME;
	else
		nstime->secs = tempstmp; /* 0 */

	nstime->nsecs = (int)(tvb_get_ntohl(tvb, offset+4)/(NTP_FLOAT_DENOM/1000000000.0));
}

/*FUNCTION:------------------------------------------------------
 *  NAME
 *      round_nsec
 *  DESCRIPTION
 *      Round integer value of nanosecond.
 *  PARAMETERS
 *      gint nsec - input value to round.
 *      gint by   - round by to number.
 *  RETURNS
 *      gint - rounded value.
 *---------------------------------------------------------------
 */
gint round_nsec(gint nsec, gint by)
{
    gint total = nsec / by;
    gint sup = ((nsec % by) >= (by / 2)) ? 1 : 0;
    return (total + sup) * by;
}
/*FUNCTION:------------------------------------------------------
 *  NAME
 *      dissect_zepv3
 *  DESCRIPTION
 *      IEEE 802.15.4 packet dissection routine for Wireshark.
 *  PARAMETERS
 *      tvbuff_t *tvb       - pointer to buffer containing raw packet.
 *      packet_info *pinfo  - pointer to packet information fields
 *      proto_tree *tree    - pointer to data tree Wireshark uses to display packet.
 *  RETURNS
 *      void
 *---------------------------------------------------------------
 */
static void dissect_zepv3(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
    tvbuff_t            *next_tvb;
    //proto_item          *proto_root, *pi;
    proto_item          *proto_root;
    proto_tree          *zepv3_tree;
    guint8              ieee_packet_len;
    guint8              zepv3_header_len;
    t_zepv3_info            zepv3_data;
    proto_item          *pi;
//    proto_tree   *zepv3_time_tree;
    
    conversation_t *conversation;
    zepv3_conv_info_t *zepv3_info;
    zepv3_transaction_t *zepv3_trans;
    
    nstime_t    temp_abs_time;
    
    zepv3_proto_data_t *proto_data; 
  

    dissector_handle_t  next_dissector;

    /*  Determine whether this is a Q51/IEEE 802.15.4, Open Sniffer packet or not */
    if(strcmp(tvb_get_ephemeral_string(tvb, 0, 2), ZEP_PREAMBLE)){
        /*  This is not a Q51/ZigBee sniffer packet */
        call_dissector(data_handle, tvb, pinfo, tree);
        return;
    }

    memset(&zepv3_data, 0, sizeof(zepv3_data)); /* Zero all zep_data fields. */

    /*  Extract the protocol version from the ZEP header. */
    zepv3_data.version = tvb_get_guint8(tvb, 2);
    if (zepv3_data.version == 1) {
        /* Type indicates a ZEP_v1 packet. */

        zepv3_header_len = ZEP_V1_HEADER_LEN;
        zepv3_data.type = 0;
        zepv3_data.channel_id = tvb_get_guint8(tvb, 3);
        zepv3_data.device_id = tvb_get_ntohs(tvb, 4);
        zepv3_data.lqi_mode = tvb_get_guint8(tvb, 6)?1:0;
        zepv3_data.lqi = tvb_get_guint8(tvb, 7);
        ieee_packet_len = (tvb_get_guint8(tvb, ZEP_V1_HEADER_LEN - 1) & ZEP_LENGTH_MASK);
    }
    else if(zepv3_data.version == 3){
    
        /*ZEP v3 native for Open Sniffer*/        
        zepv3_data.type = tvb_get_guint8(tvb, 3);
        if (zepv3_data.type == ZEP_V3_TYPE_ACK) {
            /* ZEP Ack has only the seqno. */
            zepv3_header_len = ZEP_V3_ACK_LEN;
            zepv3_data.seqno = tvb_get_ntohl(tvb, 4);
            ieee_packet_len = 0;
        }
        else {
            /* Although, only type 1 corresponds to data, if another value is present, assume it is dissected the same. */
            zepv3_header_len = ZEP_V3_HEADER_LEN;
            zepv3_data.channel_id = tvb_get_guint8(tvb, 4);
            zepv3_data.device_id = tvb_get_ntohs(tvb, 5);
            zepv3_data.lqi_mode = tvb_get_guint8(tvb, 7)?1:0;
            zepv3_data.lqi = tvb_get_guint8(tvb, 8);
//TODO:                        
            //Relative timestamp
            zepv3_data.ntp_time.secs  = tvb_get_ntohl(tvb, 9);
	          zepv3_data.ntp_time.nsecs = (int)(tvb_get_ntohl(tvb, 9+4));
            
            //round nanosecond of timestamp to 6 digits
            zepv3_data.ntp_time.nsecs = round_nsec(zepv3_data.ntp_time.nsecs,1000);
            
            zepv3_data.seqno = tvb_get_ntohl(tvb, 17);
            zepv3_data.band = tvb_get_guint8(tvb, 21);
            zepv3_data.chanpage = tvb_get_guint8(tvb, 22);
            ieee_packet_len = (tvb_get_guint8(tvb, ZEP_V3_HEADER_LEN - 1) & ZEP_LENGTH_MASK);
        }
    
    
    }
    else {
        /* At the time of writing, v2 is the latest version of ZEP, assuming
         * anything higher than v2 has identical format. */

        zepv3_data.type = tvb_get_guint8(tvb, 3);
        if (zepv3_data.type == ZEP_V2_TYPE_ACK) {
            /* ZEP Ack has only the seqno. */
            zepv3_header_len = ZEP_V2_ACK_LEN;
            zepv3_data.seqno = tvb_get_ntohl(tvb, 4);
            ieee_packet_len = 0;
        }
        else {
            /* Although, only type 1 corresponds to data, if another value is present, assume it is dissected the same. */
            zepv3_header_len = ZEP_V2_HEADER_LEN;
            zepv3_data.channel_id = tvb_get_guint8(tvb, 4);
            zepv3_data.device_id = tvb_get_ntohs(tvb, 5);
            zepv3_data.lqi_mode = tvb_get_guint8(tvb, 7)?1:0;
            zepv3_data.lqi = tvb_get_guint8(tvb, 8);             
            ntp_to_nstime(tvb, 9, &(zepv3_data.ntp_time));            
            zepv3_data.seqno = tvb_get_ntohl(tvb, 17);
            ieee_packet_len = (tvb_get_guint8(tvb, ZEP_V2_HEADER_LEN - 1) & ZEP_LENGTH_MASK);
        }
    }

#if 0
/*??dat*/
    if (zepv3_data.ntp_time.secs && zepv3_data.ntp_time.nsecs) {
        pinfo->fd->abs_ts = zepv3_data.ntp_time;
    }
#endif

    if(ieee_packet_len < tvb_length(tvb)-zepv3_header_len){
        /* Packet's length is mis-reported, abort dissection */
        call_dissector(data_handle, tvb, pinfo, tree);
        return;
    }

    /*  Enter name info protocol field */
    if(check_col(pinfo->cinfo, COL_PROTOCOL)){
        //col_set_str(pinfo->cinfo, COL_PROTOCOL, (zepv3_data.version==1)?"ZEP":"ZEPv2");
        col_add_fstr(pinfo->cinfo, COL_PROTOCOL, "ZEPv%d",zepv3_data.version);
    }

    /*  Enter name info protocol field */
    if(check_col(pinfo->cinfo, COL_INFO)){
        col_clear(pinfo->cinfo, COL_INFO);
        if (!((zepv3_data.version>=2) && (zepv3_data.type==ZEP_V2_TYPE_ACK))) col_add_fstr(pinfo->cinfo, COL_INFO, "Encapsulated ZigBee Packet [Channel]=%i [Length]=%i", zepv3_data.channel_id, ieee_packet_len);
        else col_add_fstr(pinfo->cinfo, COL_INFO, "Ack, Sequence Number: %i", zepv3_data.seqno);
    }
    
    //--------------------------------  
    
    if(zepv3_data.version == 3){
    /* Available only in ZEP v3 */
    
              /* only TRUE first time we process this packet*/              
          if (!pinfo->fd->flags.visited) {
              
              	if((conversation = find_conversation(pinfo->fd->num, &pinfo->src, &pinfo->dst,
              				                               pinfo->ptype, pinfo->srcport,
              				                               pinfo->destport, 0)) == NULL)
                {
      
                          		// No, this is a new conversation. 
                              // First captured dissector 
                          		conversation = conversation_new(pinfo->fd->num, &pinfo->src,
                          					                          &pinfo->dst, pinfo->ptype,
                          					                          pinfo->srcport, pinfo->destport, 0);
                                                              
                              zepv3_info = conversation_get_proto_data(conversation, proto_zepv3);    
                              
                              // If no data was found, create and assign data 
                              if ( zepv3_info == NULL )
                              {
                                  zepv3_info = se_alloc(sizeof(zepv3_conv_info_t));
                                  zepv3_info->pdus = se_tree_create_non_persistent(
                                      EMEM_TREE_TYPE_RED_BLACK, "zepv3_pdus");
                                  conversation_add_proto_data(conversation, proto_zepv3, zepv3_info);
                              }
                              
                                  
                              zepv3_trans = se_alloc(sizeof(zepv3_transaction_t));
                              //data save
                              zepv3_trans->ref_time = pinfo->fd->abs_ts;
                              zepv3_trans->rel_time = zepv3_data.ntp_time;  
                              zepv3_trans->prev_seqno = zepv3_data.seqno;
                              zepv3_trans->is_first = TRUE;
                              //      
                              se_tree_insert32(zepv3_info->pdus, SE_TREE_KEY, (void *)zepv3_trans);                          
                                                          
                  }else{
                            // Conversation found, get data
                            zepv3_info = conversation_get_proto_data(conversation, proto_zepv3);                      
                            zepv3_trans = se_tree_lookup32(zepv3_info->pdus, SE_TREE_KEY);
                            
                            if(zepv3_data.seqno < zepv3_trans->prev_seqno){ 
                                //Actual ZEP Sequence number lower than previous, save new data                           
                                zepv3_trans->ref_time = pinfo->fd->abs_ts;
                                zepv3_trans->rel_time = zepv3_data.ntp_time;  
                                zepv3_trans->prev_seqno = zepv3_data.seqno; 
                                zepv3_trans->is_first = TRUE;                                                
                            }
                            else{
                                //Actual ZEP Sequence number higher than previous
                                zepv3_trans->prev_seqno = zepv3_data.seqno;
                                zepv3_trans->is_first = FALSE;   
                            }                                          
                  }  
              
              
                //save data to current dissector
                proto_data = se_alloc(sizeof(zepv3_proto_data_t));
                proto_data->ref_time = zepv3_trans->ref_time;
                proto_data->rel_time = zepv3_trans->rel_time; 
                proto_data->is_first = zepv3_trans->is_first;
                
                if(zepv3_trans->is_first){
                   zepv3_trans->prev_rel_time = zepv3_data.ntp_time;
                }
                else{
                   proto_data->prev_rel_time = zepv3_trans->prev_rel_time;
                   zepv3_trans->prev_rel_time = zepv3_data.ntp_time;          
                }
                 
                p_add_proto_data(pinfo->fd, proto_zepv3,PROTO_KEY, proto_data);    
                          
            } else {
              // get saved data from current dissector
              proto_data = p_get_proto_data(pinfo->fd, proto_zepv3, PROTO_KEY);
          
            }  
            
      }               
                      
    
        //-------------------------------- 
    

    if(tree){
        /*  Create subtree for the ZEP Header */
        if (!((zepv3_data.version>=2) && (zepv3_data.type==ZEP_V2_TYPE_ACK))) {
            proto_root = proto_tree_add_protocol_format(tree, proto_zepv3, tvb, 0, zepv3_header_len, "ZigBee Encapsulation Protocol, Channel: %i, Length: %i", zepv3_data.channel_id, ieee_packet_len);
        }
        else {
            proto_root = proto_tree_add_protocol_format(tree, proto_zepv3, tvb, 0, zepv3_header_len, "ZigBee Encapsulation Protocol, Ack");
        }
        zepv3_tree = proto_item_add_subtree(proto_root, ett_zepv3);

        /*  Display the information in the subtree */
        proto_tree_add_text(zepv3_tree, tvb, 0, 2, "Protocol ID String: EX");
        if (zepv3_data.version==1) {
            proto_tree_add_uint(zepv3_tree, hf_zepv3_version, tvb, 2, 1, zepv3_data.version);
            proto_tree_add_uint(zepv3_tree, hf_zepv3_channel_id, tvb, 3, 1, zepv3_data.channel_id);
            proto_tree_add_uint(zepv3_tree, hf_zepv3_device_id, tvb, 4, 2, zepv3_data.device_id);
            proto_tree_add_boolean_format(zepv3_tree, hf_zepv3_lqi_mode, tvb, 6, 1, zepv3_data.lqi_mode, "LQI/CRC Mode: %s", zepv3_data.lqi_mode?"CRC":"LQI");
            if(!(zepv3_data.lqi_mode)){
                proto_tree_add_uint(zepv3_tree, hf_zepv3_lqi, tvb, 7, 1, zepv3_data.lqi);
            }
            proto_tree_add_text(zepv3_tree, tvb, 7+((zepv3_data.lqi_mode)?0:1), 7+((zepv3_data.lqi_mode)?1:0), "Reserved Fields");
        
        }else if (zepv3_data.version==3){
        
            /*
            *
            * ZEP version 3 
            *
            */                                                            
        
            proto_tree_add_uint(zepv3_tree, hf_zepv3_version, tvb, 2, 1, zepv3_data.version);
            if (zepv3_data.type == ZEP_V2_TYPE_ACK) {
                proto_tree_add_uint_format(zepv3_tree, hf_zepv3_type, tvb, 3, 1, zepv3_data.type, "Type: %i (Ack)", ZEP_V2_TYPE_ACK);
                proto_tree_add_uint(zepv3_tree, hf_zepv3_seqno, tvb, 4, 4, zepv3_data.seqno);
            }
            else {
                proto_tree_add_uint_format(zepv3_tree, hf_zepv3_type, tvb, 3, 1, zepv3_data.type, "Type: %i (%s)", zepv3_data.type, (zepv3_data.type==ZEP_V2_TYPE_DATA)?"Data":"Reserved");
                proto_tree_add_uint(zepv3_tree, hf_zepv3_channel_id, tvb, 4, 1, zepv3_data.channel_id);
                proto_tree_add_uint(zepv3_tree, hf_zepv3_device_id, tvb, 5, 2, zepv3_data.device_id);
                proto_tree_add_boolean_format(zepv3_tree, hf_zepv3_lqi_mode, tvb, 7, 1, zepv3_data.lqi_mode, "LQI/CRC Mode: %s", zepv3_data.lqi_mode?"CRC":"LQI");
                if(!(zepv3_data.lqi_mode)){
                    proto_tree_add_uint(zepv3_tree, hf_zepv3_lqi, tvb, 8, 1, zepv3_data.lqi);
                }
                //-----------------------------------------------------
          
                proto_tree_add_time(zepv3_tree, hf_zepv3_snifftimestamp, tvb, 9, 8, &(zepv3_data.ntp_time));
        
                //
                //Compute absolute timestamp
                //               
                nstime_delta(&temp_abs_time,&zepv3_data.ntp_time,&proto_data->rel_time);
                proto_tree_add_time(zepv3_tree, hf_zepv3_reltimestamp, tvb, 9, 8, &(temp_abs_time));
                
                nstime_sum(&temp_abs_time,&proto_data->ref_time,&temp_abs_time);     
//TODO               
                //proto_tree_add_time(zepv3_tree, hf_zepv3_abstimestamp, tvb, 9, 8, &(temp_abs_time));
                
                proto_tree_add_time_format(zepv3_tree, hf_zepv3_timestamp, tvb, 9, 8, &(temp_abs_time),
                          "Absolute Timestamp: %s",abs_time_to_str(&(temp_abs_time), ABSOLUTE_TIME_LOCAL, TRUE));
                
                
//                pi = proto_tree_add_time(zepv3_tree, hf_zepv3_abstimestamp, tvb, 9, 8, &(temp_abs_time));                   
//                zepv3_time_tree = proto_item_add_subtree(pi, ett_time_zepv3);                                              
//                proto_tree_add_time(zepv3_time_tree, hf_zepv3_refabstime, tvb, 0, 0, &(proto_data->ref_time));

                if(proto_data->is_first){
                  //proto_tree_add_text(zepv3_tree, tvb, 0, 0, "Differential Timestamp: Unknown (This is first packet)");
                  nstime_set_zero(&temp_abs_time);
                  pi = proto_tree_add_time(zepv3_tree, hf_zepv3_diftimestamp, tvb, 0, 0, &(temp_abs_time));
                  proto_item_append_text(pi, " (This is first packet)");
                }else{
                
                  nstime_delta(&temp_abs_time,&zepv3_data.ntp_time,&proto_data->prev_rel_time);
                  proto_tree_add_time(zepv3_tree, hf_zepv3_diftimestamp, tvb, 0, 0, &(temp_abs_time));
                
                }
                
                  
                //----------------------------------------------------                                
                proto_tree_add_uint(zepv3_tree,hf_zepv3_seqno, tvb, 17, 4, zepv3_data.seqno);
                
                proto_tree_add_uint(zepv3_tree, hf_zepv3_band, tvb, 21, 1, zepv3_data.band);
                
              
                if(zepv3_data.chanpage == 255){
                  proto_tree_add_uint_format_value(zepv3_tree, hf_zepv3_chanpage, tvb, 22,1, zepv3_data.chanpage,"Not Defined (%d)",zepv3_data.chanpage);
                }else{
                  proto_tree_add_uint(zepv3_tree, hf_zepv3_chanpage, tvb, 22, 1, zepv3_data.chanpage);                
                }
      
            }     
        
        
        }        
        else {
            proto_tree_add_uint(zepv3_tree, hf_zepv3_version, tvb, 2, 1, zepv3_data.version);
            if (zepv3_data.type == ZEP_V2_TYPE_ACK) {
                proto_tree_add_uint_format(zepv3_tree, hf_zepv3_type, tvb, 3, 1, zepv3_data.type, "Type: %i (Ack)", ZEP_V2_TYPE_ACK);
                proto_tree_add_uint(zepv3_tree, hf_zepv3_seqno, tvb, 4, 4, zepv3_data.seqno);
            }
            else {
                proto_tree_add_uint_format(zepv3_tree, hf_zepv3_type, tvb, 3, 1, zepv3_data.type, "Type: %i (%s)", zepv3_data.type, (zepv3_data.type==ZEP_V2_TYPE_DATA)?"Data":"Reserved");
                proto_tree_add_uint(zepv3_tree, hf_zepv3_channel_id, tvb, 4, 1, zepv3_data.channel_id);
                proto_tree_add_uint(zepv3_tree, hf_zepv3_device_id, tvb, 5, 2, zepv3_data.device_id);
                proto_tree_add_boolean_format(zepv3_tree, hf_zepv3_lqi_mode, tvb, 7, 1, zepv3_data.lqi_mode, "LQI/CRC Mode: %s", zepv3_data.lqi_mode?"CRC":"LQI");
                if(!(zepv3_data.lqi_mode)){
                    proto_tree_add_uint(zepv3_tree, hf_zepv3_lqi, tvb, 8, 1, zepv3_data.lqi);
                }
                pi = proto_tree_add_time(zepv3_tree, hf_zepv3_timestamp, tvb, 9, 8, &(zepv3_data.ntp_time));
                proto_item_append_text(pi, " (%ld.%09ds)", (long)zepv3_data.ntp_time.secs, zepv3_data.ntp_time.nsecs);
                proto_tree_add_uint(zepv3_tree, hf_zepv3_seqno, tvb, 17, 4, zepv3_data.seqno);
            }
            
        }
        if (!((zepv3_data.version==2) && (zepv3_data.type==ZEP_V2_TYPE_ACK))) proto_tree_add_uint_format(zepv3_tree, hf_zepv3_ieee_length, tvb, zepv3_header_len - 1, 1, ieee_packet_len, "Length: %i %s", ieee_packet_len, (ieee_packet_len==1)?"Byte":"Bytes");
    }

    /* Determine which dissector to call next. */
    if (zepv3_data.lqi_mode) {
        /* CRC present, use standard IEEE dissector. */
        next_dissector = ieee802154_handle;
    }
    else {
        /* ChipCon compliant FCS present. */
        next_dissector = ieee802154_ccfcs_handle;
    }
    if (!next_dissector) {
        /* IEEE 802.15.4 dissectors couldn't be found. */
        next_dissector = data_handle;
    }

    /*  Call the IEEE 802.15.4 dissector */
    if (!((zepv3_data.version>=2) && (zepv3_data.type==ZEP_V2_TYPE_ACK))) {
        next_tvb = tvb_new_subset(tvb, zepv3_header_len, ieee_packet_len, ieee_packet_len);
        call_dissector(next_dissector, next_tvb, pinfo, tree);
    }
} /* dissect_ieee802_15_4 */

/*FUNCTION:------------------------------------------------------
 *  NAME
 *      proto_register_zepv3
 *  DESCRIPTION
 *      IEEE 802.15.4 protocol registration routine.
 *  PARAMETERS
 *      none
 *  RETURNS
 *      void
 *---------------------------------------------------------------
 */
void proto_register_zepv3(void)
{
    module_t *zepv3_module;

    static hf_register_info hf[] = {
        { &hf_zepv3_version,
        { "Protocol Version",           "zepv3.version", FT_UINT8, BASE_DEC, NULL, 0x0,
            "The version of the sniffer.", HFILL }},

        { &hf_zepv3_type,
        { "Type",                       "zepv3.type", FT_UINT8, BASE_DEC, NULL, 0x0,
            NULL, HFILL }},


        { &hf_zepv3_channel_id,
        { "Channel ID",                 "zepv3.channel_id", FT_UINT8, BASE_DEC, NULL, 0x0,
            "The logical channel on which this packet was detected.", HFILL }},

        { &hf_zepv3_device_id,
        { "Device ID",                  "zepv3.device_id", FT_UINT16, BASE_DEC, NULL, 0x0,
            "The ID of the device that detected this packet.", HFILL }},

        { &hf_zepv3_lqi_mode,
        { "LQI/CRC Mode",               "zepv3.lqi_mode", FT_BOOLEAN, BASE_NONE, NULL, 0x0,
            "Determines what format the last two bytes of the MAC frame use.", HFILL }},

        { &hf_zepv3_lqi,
        { "Link Quality Indication",    "zepv3.lqi", FT_UINT8, BASE_DEC, NULL, 0x0,
            NULL, HFILL }},
            
        { &hf_zepv3_timestamp,
        { "Timestamp",                  "zepv3.time", FT_ABSOLUTE_TIME, ABSOLUTE_TIME_LOCAL, NULL, 0x0,
            NULL, HFILL }},

        { &hf_zepv3_snifftimestamp,
        { "Sniffer Timestamp",                  "zepv3.snifftime", FT_RELATIVE_TIME, BASE_NONE, NULL, 0x0,
            NULL, HFILL }},
            
        { &hf_zepv3_reltimestamp,
        { "Relative Timestamp",                  "zepv3.reltime", FT_RELATIVE_TIME, BASE_NONE, NULL, 0x0,
            NULL, HFILL }},            
             
        { &hf_zepv3_diftimestamp,
        { "Differential Timestamp",                  "zepv3.diftime", FT_RELATIVE_TIME, BASE_NONE, NULL, 0x0,
            "Differential timestamp between previous and current packet.", HFILL }},   
/*            
         { &hf_zepv3_refabstime,
        { "Reference Absolute Timestamp",         "zepv3.refatime", FT_ABSOLUTE_TIME, ABSOLUTE_TIME_LOCAL, NULL, 0x0,
            NULL, HFILL }},
*/            
        { &hf_zepv3_seqno,
        { "Sequence Number",            "zepv3.seqno", FT_UINT16, BASE_DEC, NULL, 0x0,
            NULL, HFILL }},

        { &hf_zepv3_band,
        { "Frequency band",            "zepv3.band", FT_UINT8, BASE_DEC, VALS(bandstrings), 0x0,
            "Frequency band.", HFILL }},

        { &hf_zepv3_chanpage,
        { "Channel page",            "zepv3.chanpage", FT_UINT8, BASE_DEC, NULL, 0x0,
            "Channel page.", HFILL }},            
            
          
        { &hf_zepv3_ieee_length,
        { "Length",              "zepv3.length", FT_UINT8, BASE_DEC, NULL, 0x0,
            "The length (in bytes) of the encapsulated IEEE 802.15.4 MAC frame.", HFILL }},
    };

    static gint *ett[] = {
        &ett_zepv3//,
//        &ett_time_zepv3
    };

    /*  Register protocol name and description. */
    //proto_zepv3 = proto_register_protocol("ZigBee Encapsulation Protocol - absolute timestamp", "ZEP-abs", "zepv3");
    proto_zepv3 = proto_register_protocol("ZigBee Encapsulation Protocol (v3)", "ZEPv3", "zepv3");

    /*  Register header fields and subtrees. */
    proto_register_field_array(proto_zepv3, hf, array_length(hf));
    proto_register_subtree_array(ett, array_length(ett));

    /*  Register preferences module */
    zepv3_module = prefs_register_protocol(proto_zepv3, proto_reg_handoff_zepv3);

    /*  Register preferences */
    prefs_register_uint_preference(zepv3_module, "udp.port", "ZEP UDP port",
                 "Set the port for ZEP Protocol\n"
                 "Default port is 17754",
                 10, &gPREF_zepv3_udp_port);

    /*  Register dissector with Wireshark. */
    register_dissector("zepv3", dissect_zepv3, proto_zepv3);
} /* proto_register_zepv3 */

/*FUNCTION:------------------------------------------------------
 *  NAME
 *      proto_reg_handoff_zepv3
 *  DESCRIPTION
 *      Registers the zigbee dissector with Wireshark.
 *      Will be called every time 'apply' is pressed in the preferences menu.
 *  PARAMETERS
 *      none
 *  RETURNS
 *      void
 *---------------------------------------------------------------
 */
void proto_reg_handoff_zepv3(void)
{
    static dissector_handle_t  zepv3_handle;
    static int                 lastPort;
    static gboolean            inited = FALSE;

    if ( !inited) {
        dissector_handle_t h;
        /* Get dissector handles. */
        if ( !(h = find_dissector("wpan")) ) { /* Try use built-in 802.15.4 disector */
            h = find_dissector("ieee802154");  /* otherwise use older 802.15.4 plugin disector */
        }
        ieee802154_handle = h;
        if ( !(h = find_dissector("wpan_cc24xx")) ) { /* Try use built-in 802.15.4 (Chipcon) disector */
            h = find_dissector("ieee802154_ccfcs");   /* otherwise use older 802.15.4 (Chipcon) plugin disector */
        }
        ieee802154_ccfcs_handle = h;
        zepv3_handle = find_dissector("zepv3");
        data_handle = find_dissector("data");
        inited = TRUE;
    } else {
        /* If we were already registered, de-register our dissector
         * to free the port. */
        dissector_delete_uint("udp.port", lastPort, zepv3_handle);
    }

    /* Register our dissector. */
    dissector_add_uint("udp.port", gPREF_zepv3_udp_port, zepv3_handle);
    lastPort = gPREF_zepv3_udp_port;
} /* proto_reg_handoff_zepv3 */

